open Cduce_types
open Encodings

let print ppf v =
  Value.print ppf {: v :}

let to_string v =
  let b = Buffer.create 255 in
  let ppf = Format.formatter_of_buffer b in
  print ppf v;
  Format.pp_print_flush ppf ();
  Buffer.contents b

module Namespaces = struct
  type t = (string * string) list

  let get v = match {: v :} with
    | Value.XmlNs (_,_,_,ns) -> 
	List.map 
	  (fun (a,b) -> Utf8.get_str a, Utf8.get_str b) (Ns.get_table ns)
    | _ -> raise Not_found 

  let mk_table ns =
    Ns.mk_table (List.map (fun (a,b) -> Utf8.mk a, Utf8.mk b) ns)

  let set v ns =  match {: v :} with
    | Value.XmlNs (v1,v2,v3,_) | Value.Xml (v1,v2,v3) -> 
	Obj.magic (Value.XmlNs (v1,v2,v3,mk_table ns))
    | _ -> assert false
end


module Print = struct
  open Value
  open Ident
  exception Error of string * {{ Any }}

  let serialize ~start_elem ~end_elem ~empty_elem ~text ?(ns=[]) v =
    let ns = Namespaces.mk_table ns in
    let v = {: v :} in
    let printer = Ns.Printer.printer ns in

    let to_str = Encodings.Utf8.get_str in

    let mk_attrs =
      List.map (fun (n,v) ->
		  Ns.Printer.attr printer n, to_str v) in
    let mk_xmlns = 
      List.map (fun (pr,ns) ->
		  let pr = to_str pr in
		  let pr = if pr = "" then "xmlns" else ("xmlns:" ^ pr) in
		  pr, to_str (Ns.value ns)) in
    let element_start n xmlns attrs = 
      start_elem (Ns.Printer.tag printer n) (mk_xmlns xmlns @ mk_attrs attrs)
    and empty_element n xmlns attrs = 
      empty_elem (Ns.Printer.tag printer n)  (mk_xmlns xmlns @ mk_attrs attrs)
    and element_end n = 
      end_elem (Ns.Printer.attr printer n)
    in
    
    let rec register_elt = function
      | Xml (Atom tag, Record attrs, content) 
      | XmlNs (Atom tag, Record attrs, content, _) ->
	  List.iter
	    (fun (n,_) -> Ns.Printer.register_attr printer (LabelPool.value n))
	    (LabelMap.get attrs);
	  Ns.Printer.register_tag printer (Atoms.V.value tag);
	  register_content content
      | _ -> ()
    and register_content = function
      | String_utf8 (_,_,_,q)
      | String_latin1 (_,_,_,q) -> register_content q
      | Pair (x, q) -> register_elt x; register_content q
      | Concat (x,y) -> register_content x; register_content y
      | _ -> () 
    in
    register_elt v;
    
    let rec print_elt xmlns = function
      | Xml (Atom tag, Record attrs, content)
      | XmlNs (Atom tag, Record attrs, content, _) ->
	  let tag = Atoms.V.value tag in
	  let attrs = 
	    LabelMap.mapi_to_list 
	      (fun n v -> 
		 if is_str v then
                   let (s,_) = get_string_utf8 v in
		   (LabelPool.value n, s)
		 else
		   raise (Error ("Attribute value is not a string", {{{:v:}}}))
	      ) attrs in
	  if Value.equal content Value.nil then
	    empty_element tag xmlns attrs
	  else (
	    element_start tag xmlns attrs;
	    print_content content;
	    element_end tag
	  )
      | v -> raise (Error ("Content value is not valid XML", {{{:v:}}}))
    and print_content v =
      let (s,q) = get_string_utf8 v in
      text (to_str s);
      match normalize q with
	| Pair ((Xml _ | XmlNs _) as x, q) -> print_elt [] x; print_content q
	| Atom a when a = Sequence.nil_atom -> ()
	| v -> raise (Error ("Content value is not valid XML", {{{:v:}}}))
    in
    print_elt (Ns.Printer.prefixes printer) v


  let data f s =
    let rec aux b i =
      if i = String.length s then f (String.sub s b (i-b))
      else match s.[i] with
	| '&' -> f (String.sub s b (i-b)); f "&amp;"; aux (succ i) (succ i)
	| '<' -> f (String.sub s b (i-b)); f "&lt;"; aux (succ i) (succ i)
	| '>' -> f (String.sub s b (i-b)); f "&gt;"; aux (succ i) (succ i)
	| '"' -> f (String.sub s b (i-b)); f "&quot;"; aux (succ i) (succ i)
	| _ -> aux b (succ i)
    in
    aux 0 0

  let escape f s =
    let rec aux b i =
      if i = String.length s then f (String.sub s b (i-b))
      else match s.[i] with
	| '&' -> f (String.sub s b (i-b)); f "&amp;"; aux (succ i) (succ i)
	| '<' -> f (String.sub s b (i-b)); f "&lt;"; aux (succ i) (succ i)
	| '>' -> f (String.sub s b (i-b)); f "&gt;"; aux (succ i) (succ i)
	| '"' -> f (String.sub s b (i-b)); f "&quot;"; aux (succ i) (succ i)
	| _ -> aux b (succ i) in
    aux 0 0

  let print_xml f ?(data=escape f) ?ns v =
    let open_markup tag attrs =
      f ("<" ^ tag);
      List.iter (fun (n,v) -> f (" " ^ n ^ "=\""); data v; f "\"") attrs
    in
    serialize
      ~start_elem:(fun tag attrs -> open_markup tag attrs; f ">")
      ~end_elem:(fun tag -> f ("</" ^ tag ^ ">"))
      ~empty_elem:(fun tag attrs -> open_markup tag attrs; f "/>")
      ~text:data
      ?ns
      v
end


module Load = struct
  open Ident
  open Value
  exception Error of string
  type anyxml = {{ <_ ..>[ (Char | anyxml)* ] }}

  type stk = 
    | Element of Value.t * stk
    | Start of Ns.table * Ns.qname * (Ns.qname * Utf8.t) list * Ns.table * stk
    | String of string * stk
    | Empty

  type t = { 
    mutable stack : stk; 
    mutable ns_table : Ns.table;

    mutable buffer : string;
    mutable pos : int;
    mutable length : int;
    
    keep_ns: bool
  }

  let make ?(ns=false) () =
    { stack = Empty;
      ns_table = Ns.empty_table;
      buffer = String.create 1024;
      pos = 0;
      length = 1024;
      keep_ns = ns }
      
  let get loader =
    match loader.stack with
      | Element (x,Empty) -> Obj.magic x
      | _ -> raise (Error "No XML stream to parse")

  let resize loader n  =
    let new_len = loader.length * 2 + n in
    let new_buf = String.create new_len in
    String.unsafe_blit loader.buffer 0 new_buf 0 loader.pos;
    loader.buffer <- new_buf;
    loader.length <- new_len
      
  let text loader s =
    let len = String.length s in
    let new_pos = loader.pos + len in
    if new_pos > loader.length then resize loader len;
    String.unsafe_blit s 0 loader.buffer loader.pos len;
    loader.pos <- new_pos
      
  let rec only_ws s i =
    (i = 0) ||
      (let i = pred i in match (String.unsafe_get s i) with
	 | ' ' | '\t' | '\n' | '\r' -> only_ws s i
	 | _ -> false) 
      
      
  let string s q =
    let s = Utf8.mk s in
    String_utf8 (Utf8.start_index s,Utf8.end_index s, s, q)
      

  let attrib att = 
    let att = List.map (fun (n,v) -> LabelPool.mk n, string_utf8 v) att in
    LabelMap.from_list 
      (fun _ _ -> raise (Error "Invalid XML document: uniqueness of attributes")) att
      
  let elem keep_ns ns (tag_ns,tag) att child =
    if keep_ns then
      XmlNs (Atom (Atoms.V.mk tag_ns tag), Record (attrib att), child, ns)
    else
      Xml (Atom (Atoms.V.mk tag_ns tag), Record (attrib att), child)
	
  let rec create_elt loader accu = function
    | String (s,st) -> create_elt loader (string s accu) st
    | Element (x,st) -> create_elt loader (Pair (x,accu)) st
    | Start (ns,name,att,old_table,st) -> 
	loader.stack <- Element (elem loader.keep_ns ns name att accu, st);
	loader.ns_table <- old_table
    | Empty -> assert false
	
  let flush_txt loader =
    if not (only_ws loader.buffer loader.pos) then 
      loader.stack <- 
	String (String.sub loader.buffer 0 loader.pos, loader.stack); 
    loader.pos <- 0


  let start_elem loader name att =
    flush_txt loader;
    let (table,name,att) = Ns.process_start_tag loader.ns_table name att in
    loader.stack <- Start (table,name,att,loader.ns_table, loader.stack);
    loader.ns_table <- table
      
  let end_elem loader _ =
    let accu =
      if only_ws loader.buffer loader.pos 
      then nil 
      else string (String.sub loader.buffer 0 loader.pos) nil in
    loader.pos <- 0; 
    create_elt loader accu loader.stack

  let sub loader v =
    flush_txt loader;
    loader.stack <- Element ({:v:}, loader.stack)
end

module Utf8 = struct
  type t = utf8

  let to_string_utf8 x = Utf8.get_str (Obj.magic x)
  let of_string_utf8 x = 
    match Utf8.mk_check x with Some x -> (Obj.magic x)
      | None -> failwith "of_string_utf8"

  let to_string_latin1 x = Utf8.to_string (Obj.magic x)
  let of_string_latin1 x = Obj.magic (Utf8.mk_latin1 x)
end


module Types = struct
  type t = Cduce_types.Types.t
  let print = Cduce_types.Types.Print.print
end

module Check = struct
  type t = Cduce_types.Explain.t
  exception Failed = Cduce_types.Explain.Failed
  let get x = Obj.magic x
  let print = Cduce_types.Explain.print
  let to_string = Cduce_types.Explain.to_string
  let simplify = Cduce_types.Explain.simplify
end
